; i386 specific definitions
; ===========================================================================

; ---------------------------------------------------------------------------
; Useful Constants
; ---------------------------------------------------------------------------

(define FLAG_CF		#x00000001) ;	Carry
(define FLAG_SET	#x00000002) ;	????
(define FLAG_PF		#x00000004) ;	Parity
(define FLAG_AF		#x00000010) ;	Auxiliary Carry for BCD
(define FLAG_ZF		#x00000040) ;	Zero
(define FLAG_SF		#x00000080) ;	Sign
(define FLAG_TF		#x00000100) ;	Trap
(define FLAG_IF		#x00000200) ;	Interrupt 0:Disable 1:Enable
(define FLAG_DF		#x00000400) ;	Direction 0:Upward  1:Downward
(define FLAG_OF		#x00000800) ;	Overflow
(define FLAG_IOPL	#x00003000) ;	IO Priviledge requirement mask
(define FLAG_IOPL1	#x00001000) ;	IO Priviledge requirement 1
(define FLAG_IOPL2	#x00002000) ;	IO Priviledge requirement 2
(define FLAG_IOPL3	#x00003000) ;	IO Priviledge requirement 3
(define FLAG_NT		#x00004000) ;	Nested Task
(define FLAG_RF		#x00010000) ;	Resume debug
(define FLAG_VM		#x00020000) ;	V86 Mode
(define FLAG_AC		#x00040000) ;	Alignment Check
(define FLAG_ID		#x00200000) ;	CPUID available

(define CR0_PE	#x00000001) ;	Protected mode Enable
(define CR0_MP	#x00000002) ;	Math copro Present
(define CR0_EM	#x00000004) ;	Emulate Math copro
(define CR0_TS	#x00000008) ;	Task Switched
(define CR0_ET	#x00000010) ;	Extension Type
(define CR0_NE	#x00000020) ;	Numeric Error
(define CR0_WP	#x00010000) ;	Write Protect
(define CR0_AM	#x00020000) ;	Alignment Mask
(define CR0_NW	#x20000000) ;	No Write-back
(define CR0_CD	#x40000000) ;	Cache Disable
(define CR0_PG	#x80000000) ;	Pageing

(define CR3_PWT	#x00000004) ;	Page Write Transparency
(define CR3_PCD	#x00000010) ;	Page Cache Disable


; .........................
; ASCII codes
; .........................
(define LF	#x0A)
(define CR	#x0D)
(define Escape	#x1B)

; .........................
; Keyboard scancodes
; .........................
(define K_Home	#x47)
(define K_Up	#x48)
(define K_PgUp	#x49)
(define K_Lft	#x4B)
(define K_Rgt	#x4D)
(define K_EndC	#x4F)
(define K_Dn	#x50)
(define K_PgDn	#x51)
(define K_InsC	#x52)
(define K_Del	#x53)


; .........................
; Architecture-dependent information
; .........................

(define BYTE-ORDER 1234) ;(little endian)
(define PAGESIZE 4096)
(define PAGEBITS 12)

; ---------------------------------------------------------------------------
; Assembly control information
; ---------------------------------------------------------------------------

; Curent mode of the CPU. Can be changed with set!
(define mode32 #f)

; Choose the appropriate version of the register according to the current mode.
(define (r?? r) ((if mode32 r32 r16) r))

; ---------------------------------------------------------------------------
; Instruction set enhancement macros
; ---------------------------------------------------------------------------
					   
; Determines if this is a register on which most operations can be done
(define (general-reg? r) 
  (and (reg? r) (member (reg-symbol r)
			'(al ah ax eax bh bl bx ebx
			  ch cl cx ecx dh dl dx edx
			  sp bp si di esp ebp esi edi))))

(define (valid-addressing? op) (******))
(define (valid-operand? op) (or (general-reg? op) (valid-addressing? op)))
					   
; Clear all listed operands
(define (clr . ops) 
  (apply seq (map (lambda (op)
		    (if (general-reg? op)
			(inst 'xor op op)
			(inst 'mov (assert valid-addressing? op) 0)))
		  ops)))

; Set all bits of listed operands
(define (setf . ops) 
  (apply seq (map (lambda (op)
		    (if (general-reg? op)
			(inst 'or op op)
			(inst 'mov (assert valid-addressing? op) -1)))
		  ops)))
  
; Changes byte order of a 32 bit general register
(define (bswaperx r)
  (assert general-reg? r)
  (seq (inst 'xchg (r8l r) (r8h r))
       (inst 'rol  (r32 r) 16) 
       (inst 'xchg (r8l r) (r8h r))))
       
(define (flush-instr)
  (let ((l (gen-label)))
    (seq (inst 'jmp l)
	 l)))

(define (die)
  (let ((l (gen-label)))
    (seq l
	 (inst 'hlt)
	 (inst 'jmp l))))

(define (pushfX) (inst (if mode32 'pushfd 'pushf)))
(define (popfX) (inst (if mode32 'popfd 'popf)))
(define (pushaX) (inst (if mode32 'pushad 'pusha)))
(define (popaX) (inst (if mode32 'popad 'popa)))
; * BEWARE !!!
; The instruction following a popa/popad MUST NOT use the base+index register
; addressing mode, or else most 386 cpus will trash eax, or even crash if eax
; is used in the addressing. That's an intel bug. Avoid it.

; Push in reverse order the list of operands
; Apart from operands, symbols a ad ax f fd fx will call push[a|f][d|X]?
(define (pushm . ops) 
  (apply seq (map (lambda (op)
		    (if (equal? op (r?? ip))
			(let ((l (gen-label)))
			  (seq (call l)
			       l))
			(if (symbol? op)
			    (case op
			      ((a) (inst 'pusha))
			      ((ad) (inst 'pushad))
			      ((ax) (pushaX))
			      ((f) (inst 'pushf))
			      ((fd) (inst 'pushfd))
			      ((fx) (pushfX)))
			    (begin (assert valid-operand? op)
				   (inst 'push op)))))
		  (reverse ops))))

; Pop in given order the list of operands (so the same operand list that
; whas used to pushm should be use to popm to undo the pushm)
(define (popm . ops) 
  (apply seq (map (lambda (op)
		    (if (equal? op (r?? ip))
			(inst 'ret)
			(if (symbol? op)
			    (case op
			      ((a) (inst 'popa))
			      ((ad) (inst 'popad))
			      ((ax) (popaX))
			      ((f) (inst 'popf))
			      ((fd) (inst 'popfd))
			      ((fx) (popfX)))
			    (begin (assert valid-operand? op)
				   (inst 'pop op)))))
		  ops)))

(define (pushing operands . program)
  (seq (pushm operands)
       (apply seq program)
       (popm operands)))

; Move r2 to r1 by pushing/popping
(define (ppmov r1 r2) (seq (pushm r2) (popm r1)))

; Move r2 to r1 through temporary reg/mem r3
(define (ttmov r1 r2 r3)
  (assert reg? r1)
  (assert reg? r2)
  (assert valid-operand? r3)
  (seq (inst 'mov r3 r2)
       (inst 'mov r1 r3)))

(define (movzil r op)
  (seq (clr (r32 r))
       (inst 'mov (r8l r) op)))

(define (movm . args)
  (if (null? args)
      (seq)
      (apply seq (map (lambda (x) 
			(assert valid-operand? (car x))
			(inst 'mov (car x) (cadr x)))
		      args))))
      

(define (SegOfsToLinear r1 r2 op1 op2. rest)
  (let ((adjust (if (null? rest) #f (car rest)))
	(update (if (or (null? rest) (null? (cdr rest))) #f (cadr rest))))
    (seq (inst 'movzx r1 (op-size word op2))
	 (inst 'movzx (r32 r2) (op-size word op1))
	 (if adjust (AddIm r2 adjust) (seq))
	 (inst 'shl r1 4)
	 (inst 'add r1 (r32 r2))
	 (if update (inst 'mov op1 r2) (seq)))))


; ---------------------------------------------------------------------------
; Structured programming 
; ---------------------------------------------------------------------------

; That's were I'm at. Some of this has already been done in
; control_structures.scm

; ---------------------------------------------------------------------------
; Console I/O
; ---------------------------------------------------------------------------

(define (PUTC . rest)
  (let ((r (if (null? rest) #f (car rest))))
    (seq (if r (inst 'mov al r) (seq))
	 (call 'putc))))

(define (CGETC . rest)
  (let ((text (if (null? rest) #f (car rest)))
	(opt (if (or (null? rest) (null? (cdr rest))) #f (cadr rest))))
    (seq (if opt (pushaX) (seq))
	 (if text (PUTS text) (seq))
	 (if opt (popaX) (seq))
	 (call 'cgetc))))
	 
; ******** There's more to translate here ********


; ---------------------------------------------------------------------------
; Library
; ---------------------------------------------------------------------------

; A function that retreives the required function from the library.
; The returned program used "mode" to know how to return and args
; as actual parameters for the formal parameters that appear in the
; library entry.
(define (lib-fetch name mode . args)
  (let ((entry (assoc name lib-alist)))
    (if entry
	(apply ((cdr entry) mode) args)
	(error "Can't find this entry in library :" name))))

(define lib-alist (list))

(define (lib-add! name fun)
  (set! lib-alist (cons (cons name fun) lib-alist)))

; A return instruction that depends on the "calling" mode.
(define (ret mode)
  (case mode
    ((inline) (seq))
    ((far) (inst 'retf))
    ((near) (inst 'ret))
    (else (error "Invalid mode: " mode))))



; .........................
; detecting CPU type
; .........................

(lib-add! 'testCPU
	  ; Procedure initially taken from Robert L. Hummel's book
	  ; on 80x86 programming, then enhanced with the help of Cedric Ware.
	  ;
	  ; AH
	  ; 01	8088 or 8086
	  ; 02	80286
	  ; 03	80386
	  ; 04	486
	  ; 05	Pentium
	  ; 06	P6
	  ;
	  ; AL
	  ; 00	No FPU
	  ; 01	8087
	  ; 02	80287
	  ; 03	80387
	  ; 04	486DX or 487SX
	  ; 05	Pentium
	  ; 06	P6
	  ;
	  ; Note: 0400 <=> 486SX w/o FPU.
	  ;
	  ; destroys all general purpose registers, including FP registers
	  (lambda (mode)
	    (lambda ()
	      (seq 
	       (inst 'mov dx #x0100) ; default answer
	       (inst 'mov ax -1)   ; non zero
	       (inst 'push ax)   ; was [NDP_STATUS],-1 ;
	       (inst 'mov bp sp)   ; save SP
	       ; on 8088/8086, bits 12 to 15 of pushed Flags value are set,
	       ; that's why we'll try to clear them.
	       (inst 'pushf)
	       (inst 'pop ax)
	       (inst 'and ah #x0f)   ;(clear most significant bits)
	       (inst 'push ax)
	       (inst 'popf)
	       (inst 'pushf)
	       (inst 'pop ax)
	       (inst 'and ah #xf0)   ;(get them back)
	       (inst 'cmp ah #xf0)
	       (inst 'jz (label 'cpuid_2))   ;if all set, it's a 8086/8088
	       ; on the 80286, on the opposite, they are cleared
	       (inst 'inc dh)
	       (inst 'pushf)
	       (inst 'pop ax)
	       (inst 'or ah #xf0)   ;(set most significant bits)
	       (inst 'push ax)
	       (inst 'popf)
	       (inst 'psuhf)
	       (inst 'pop ax)
	       (inst 'and ah #xf0)   ;(get them back)
	       (inst 'jz (label 'cpuid_2))   ;if cleared, it's a 80286
	       ; On the 80486, the EFLAGS registers has got an AC alignment bit
	       ; From late 486s on, the EFLAGS register has an ID bit 
	       ; (authorize CPUID)
	       (inst 'inc dh)
	       (inst 'cli)
	       (inst 'and sp -4) ; -4 for (not 3) I hope that's right (P.P.)
	                         ; align sp
	       (inst 'pushfd)
	       (inst 'pop eax)
	       (inst 'mov ebx eax)   ; save eflags
	       (inst 'xor eax (bitwise-or FLAG_AC FLAG_ID))
	       (inst 'push eax)
	       (inst 'popfd)
	       (inst 'pushfd)
	       (inst 'pop eax)
	       (inst 'push ebx)
	       (inst 'popfd)   ; restore AC
	       (inst 'mov sp bp)   ; restore sp
	       (inst 'xor eax ebx)   ; look if AC or ID was modified
	       (inst 'sti)
	       (inst 'jz (label 'cpuid_2))   ; no: it's a 386
	       (inst 'inc dh)
	       (inst 'test eax FLAG_ID)   ; ID modified ?
	       ;(!) I'm not sure this works like this:
	       ;perhaps instead I should test FLAG_ID...
	       (inst 'jz (label 'cpuid_2))   ; no, it's an early 486
	       ; Now, we can use the CPUID instruction to determine the
	       ; CPU type
	       (clr eax)   ; sequence smaller than mov eax,1 !
	       (inst 'inc eax)
	       (data byte #x0F)   ; These 2 lines were 1 before (P.P.)
	       (data byte #xA2)   ; I hope it still works. CPUID
	       (inst 'and ah 7)   ; CPU type in eax bits [11:8]
	       (inst 'mov dh ah)   ; stored in dh
	       ; Now, test for coprocessor:
	       (label 'cpuid_2)
	       (inst 'fninit)   ;reset status word
	       (inst 'gnstsw (index bp))
	       (inst 'cmp (op-size byte (index bp)) 0)   ;if non zero, no FPU
	       (inst 'jnz (label 'cpuid_3))
	       (inst 'fnstcw (index bp))   ;try to write control word
	       (inst 'and (op-size word (index bp)) #x103f)
	       (inst 'cmp (op-size word (index bp)) #x003f)
	       (inst 'jnz (label 'cpuid_3))
	       ; The only case when the FPU doesn't answer is a 80386+80287
	       (inst 'mov dl dh)   ;initialize FPU type
	       (inst 'cmp dh 3)   ;no 386 => ok
	       (inst 'jnz (label 'cpuid_3))
	       ; 387/287: the initialized NDP must manage infinity,
	       ; but they're the same to the 287	FLD1
	       (inst 'fldz)
	       (inst 'fdiv (reg 'st) (reg 'st))
	       (inst 'fld (reg 'st))
	       (inst 'fchs)
	       (inst 'fcompp)
	       (inst 'fstsw (index bp))
	       (inst 'mov ax (index bp))
	       (inst 'sahf)
	       (inst 'jnz (label 'cpuid_3))   ;different => 387
	       (inst 'dec dl)
	       (label 'cpuid_3)
	       (inst 'mov ax dx)
	       (ret mode)))))



; ---------------------------------------------------------------------------
; Memory management in protected mode
; ---------------------------------------------------------------------------


; see file "mem386.doc" for dox...

; .........................
; A20 line control
; .........................

; Constants to access line A20
(define KB_command_port #x64)  ;IOaddr to write to
(define KB_status_port  #x64)  ;IOaddr to read from
(define KB_data_port    #x60)  ;IOaddr to read from
(define IB_free         #x02)  ;status bitmask: is kbd controller ready ?
(define KB_command_A20  #xD1)  ;command bitmask to access A20 line.
(define set_A20_ON      #xDF)
(define set_A20_OFF     #xDD)

; .........................
; Macros to control line A20
; .........................

(lib-add! 'kb_ready
	  ; Test if kbd port is available (needed before to send it commands)
	  ; - {$1} must be a label where to jump if line not available
	  ; - {$2} must be the number of tries before to jump to {$1}
	  ; ** destroys AX,CX
	  (lambda (mode)
	    (lambda (fail-label . rest)
	      (let ((tries (if (null? rest) 100 (car rest)))
		    (retry-label (gen-label)))
		(seq (if (equal? tries (r?? cx))
			 (seq)
			 (inst 'mov  (r?? cx) tries))
		     retry-label
		     (inst 'in al KB_status_port)
		     (inst 'test al IB_free)
		     (inst 'loopne retry-label)
		     (inst 'jcxz fail-label)
		     (ret mode))))))


(lib-add! 'set_a20
	  ; (de)activivates line A20,
	  ; - {$1} is "ON" or "OFF",
          ;    or absent (then ah contains A20_ON or A20_OFF)
	  ; - (!) keyboard must be ready (use previous macro for that)
	  ; ** destroys AX
	  (lambda (mode)
	    (lambda rest
	      (let ((state (if (null? rest) #f (car rest))))
		(seq (if state
			 (inst 'mov ah (if (equal? state "ON")
					   set_A20_ON
					   set_A20_OFF))
			 (seq))
		     (inst 'mov al KB_command_A20)
		     (inst 'out KB_command_port al)
		     ; Linus does call KB_Ready again !
		     (inst 'mov al ah)
		     (inst 'out KB_command_port al)
		     ; Linus does call KB_Ready again !
		     (ret mode))))))
			       
(lib-add! 'test_a20
	  ; * tests if line A20 is activated, by seeing whether addresses are
	  ;  wrapped over 1MB.
	  ; * sets ax iff the line is activated
	  (lambda (mode)
	    (lambda ()
	      (let ((exit-label (gen-label)))
		(seq (pushm ds es 'fd)
		     (inst 'cli) ; disable interrupts
		     (clr ax)
		     (inst 'mov es ax) ; es=$0000
		     (inst 'dec ax)
		     (inst 'mov ds ax) ; ds=$FFFF
		     (inst 'mov ax (index -2)) ; verify [$10FFEE]=[$0FFEE]
		     (inst 'seg es)
		     (inst 'cmp ax (index -18))
		     (inst 'stc)
		     (inst 'jnz exit-label)
		     (inst 'inc (index -2)) ; verify [$10FFEE]=[$0FFEE]
		                            ; after modification
		     (inst 'seg es)
		     (inst 'cmp ax (index -18))
		     (inst 'mov (index -2) ax)
		     (inst 'mov ax -1)
		     (inst 'jz exit-label)
		     (inst 'inc ax)
		     exit-label
		     (popm ds es 'fd) ; enable interrupts again if needed
		     (ret mode))))))


; .........................
; NMI enabling/disabling
; .........................

(lib-add! 'disable_NMI
	  (lambda (mode)
	    (lambda ()
	      (seq (inst 'mov al #x80)
		   (inst 'out #x70 al)
		   (inst 'in al #x71)
		   (ret mode)))))

(lib-add! 'enable_NMI
	  (lambda (mode)
	    (lambda ()
	      (seq (inst 'mov al #x00)
		   (inst 'out #x70 al)
		   (inst 'in al #x71)
		   (ret mode)))))

; "...to accomplish a warm boot equivalent to Ctrl-Alt-Del, store 1234h
; in 0040h:0072h and jump to FFFFh:0000h.    For a cold boot equivalent
; to a reset, store 0000h at 0040h:0072h before jumping."

(lib-add! 'reset_computer
	  (lambda (mode)
	    (lambda ()
	      (seq (inst 'mov al #xF0)
		   (inst 'out #x64 al)))))

(define (defdescriptor p1 p2 p3 p4)
  ;  base,limit,_P_DPL_DT_Type,_G_D_0_AVL_0000
  (seq (data word (modulo p2 (<< 1 16)))
       (data word (modulo p1 (<< 1 16)))
       (data byte (modulo (>> p1 16) (<< 1 8)))
       (data byte p3)
       (data byte (+ p4 (modulo (>> p2 16) (<< 1 4))))
       (data byte (>> p1 24))))

(define (defgate p1 p2 p3 p4)
  ;  sel,ofs,000compteur,_P_DPL_DT_Type
  (seq (data word (modulo p2 (<< 1 16)))
       (data word p1)
       (data byte p3)
       (data byte p4)
       (data word (>> p2 16))))

(lib-add! 'set-descriptor 
	  ;	DESCRIPT
	  ; initialize a descriptor with
	  ; - {$1} being the base address of the descriptor to initialize
	  ; - ebx is base address of the segment
	  ; - ecx is the limit field in byte or page according to granularity
	  ; - al has got fields: P,DPL,DT,Type
	  ; - ah has got fields: G,D,AVL
	  ; * the routine saves all registers
	  (lambda (mode)
	    (lambda (r)
	      (seq (pushing (list ebx ecx)
			    (inst 'mov (index r) cx)
			    (inst 'mov (index r 2) bx)
			    (inst 'shr ecx 16)
			    (inst 'shr ebx 16)
			    (inst 'or cl ah)
			    (inst 'mov ch bh)
			    (inst 'mov bh al)
			    (inst 'mov (index r 6) cx)
			    (isnt 'mov (index r 4) bx))
		   (ret mode)))))
(lib-add! 'get-descriptor
	  ;	DESCRIPT
	  ; This macro does the converse of preceding one:
	  ; - {$1} contains base address of the descriptor
	  ; * returns in ebx the base address of the segment
	  ; * returns in ecx the limit field
	  ; * returns in al fields: P,DPL,DT,Type
	  ; * returns in ah fields: G,D,AVL
	  (lambda (mode)
	    (lambda (r)
	      (seq (inst 'mov cx (index r 6))
		   (inst 'mov bx (index r 4))
		   (inst 'mov al bh)
		   (inst 'mov bh ch)
		   (inst 'mov ah cl)
		   (inst 'and ah #xF0)
		   (inst 'and cx #x0F)
		   (inst 'shl ebx 16)
		   (inst 'shl ecx 16)
		   (inst 'mov bx (index r 2))
		   (inst 'mov cx (index r))
		   (ret mode)))))


; ---------------------------------------------------------------------------
; Constants and structures
; ---------------------------------------------------------------------------

; .........................
; Segment selectors
; .........................

(define SEL_MASK	#xFFF8)	; offset to descriptor entry in table
(define SEL_TI		#x0004)	; bit cleared:GDT set:LDT
(define SEL_RPL		#x0003)	; priviledge of segment making request


; .........................
; Segment descriptors
; .........................

(define DT_G		#x80) ; Granularity of limit 0yte 1=page
(define DT_D		#x40) ; Default size using CS/SS 0=16bit 1=32bit
(define DT_AVL		#x10) ; AVaiLable to OS (unused by CPU)
(define DT_P		#x80) ; segment Present bit
(define DT_PL0		#x00) ; descriptor PriviLedge
(define DT_PL1		#x20)
(define DT_PL2		#x40)
(define DT_PL3		#x60)
(define DT_DT		#x10) ; Descriptor Type, 1: segment, 0: special
; if DT_Segment (bit DT) is set:
(define DT_A		#x1) ; Accessed
(define DT_Data		#x0)
(define DT_Writable	#x2) ; Writable
(define DT_E		#x4) ; Extensible downward
(define DT_Code		#x8)
(define DT_R		#x2) ; Readable
(define DT_C		#x4) ; Conforming
; if DT_Segment -bit DT) is cleared:
(define DT_TSS16_A	#x1) ; available TSS 16
(define DT_LDT		#x2) ; LDT
(define DT_TSS16_B	#x3) ; busy TSS 16
(define DT_CallGate16	#x4)
(define DT_TaskGate	#x5)
(define DT_IntGate16	#x6)
(define DT_TrapGate16	#x7)
(define DT_TSS32_A	#x9) ; available TSS 32
(define DT_TSS32_B	#xB) ; busy TSS 32
(define DT_CallGate32	#xC)
(define DT_IntGate32	#xE)
(define DT_TrapGate32	#xF)
; a Trap gate requires CPL<=IOPL to be called; IF is unchanged
; an Interrupt gate clears IF; both Trap and Interrupt gates clear NT
; the counter is the number of stack arguments of a call gate.

; .........................
; Page table entries
; .........................

(define PG_P		#x01) ; Present
(define PG_RW		#x02) ; Read/Write
(define PG_US		#x04) ; User (PL3) access possible
(define PG_PWT		#x08) ; Page Write Through (486+)
(define PG_PCD		#x10) ; Page Cache Disable (486+)
(define PG_A		#x20) ; Accessed (486+)
(define PG_D		#x40) ; Dirty (undefined for page directory entries)
; bits 7,8 reserved
; bits 9-11 available to OS (typically, page locking, copy-on-write, etc)


; .........................
; Structures
; .........................

(define std-struc
  (seq
   (struc 'xmsmp ; XMS Memory move
	  ; Struc XMS_Move_Params   ;; parameters for XMS func 0Bh
	  (elem 'lengthtomove long)
	  (elem 'sourcehandle word)
	  (elem 'sourceoffset long)
	  (elem 'desthandle word)
	  (elem 'destoffset long))
   (apply struc
	  (cons 'pushad 
		(list-elems long 'edi 'esi 'ebp 'esp 'ebx 'edx 'ecx 'eax)))
   (apply struc 
	  (cons 'pusha
		(list-elems word 'di 'si 'bp 'sp 'bx 'dx 'cx 'ax)))
   (apply struc 
	  (cons 'v86int 
		(list-elems long 'edi 'esi 'ebp 'aesp 'ebx 'edx 'ecx 'eax
			         'eip 'cs 'eflags 'esp 'ss 'es 'ds 'fs 'gs)))
   (apply struc 
	  (cons 'intstack
		(list-elems long 'eip 'cs 'eflags 'esp 'ss)))
   (apply struc 
	  (cons 'tss16
		(list-elems word 'backlinkselector 
			    'sp_cpl0 'ss_cpl0 'sp_cpl1 'ss_cpl1 'sp_cpl2 
			    'ss_cpl2 'ip 'flags 'ax 'cx 'dx 'bx 'sp 'bp 
			    'si 'di 'es 'cs 'ss 'ds 'ldt)))
   (apply struc 
	  (cons 'tss32
		(append
		 (list-elems word 'backlinkselector 
			     'esp_cpl0 'ss_cpl0 'esp_cpl1 'ss_cpl1 'esp_cpl2 
			     'ss_cpl2 'cr3 'eip 'eflags 'eax 'ecx 'edx 'ebx
			     'esp 'ebp 'esi 'edi 'es 'cs 'ss 'ds 'fs 'gs 'ldt)
		 (list (elem 'TrapBit word)
		       (elem 'IO_Map_Base word)))))))

; TSS32_CR3 is reserved by Intel but everyone I know seem to use it as
; a place to store the CR3 page directory register for the given task.










; ===========================================================================
;What I'm not sure if or how to translate and other things to do:

;define({setf},{ifelse(isreg($1),1,{or $1,$1},{test $1,#-1})})
;Isn't is mov instead of test ?

;define({__unseg__},         in as86.m4
;what does that do? it was used in the push/pop stuff and I neglected it.

;movm used __mov__ which did a lot of not so obviously efficient optimization
;I used the instruction mov instead of doing __mov__.

;define({SegOfsToLinear},
;what does that do?

;newheap & allot have not been translated. Can't we use the struc instead ?

;dnl ---------------------------- Diversions ----------------------------------
;what does that do?

;dnl --------------------- Data definition ---------------------
;what does that do?

;dnl ---------------- Exportation de donnees -----------------------
;what does that do?

;function puts depends on lsa which is in exportation de donnees

; check testcpu for occurences of P.P. and verify that the modifications
; are valid

;dnl -------------- Support for install-time options dichotomy ----------------
;what does that do?

;To do, fix the immediate or not mode (# or not) for numbers in the
; translation from the internal representation to as86
; (difficult because of rol eax,16 and mov eax,#0)
; ===========================================================================
